<?php

namespace Drupal\config_enforce;

use Drupal\Component\Utility\Crypt;
use Drupal\Core\Config\ConfigImporter;
use Drupal\Core\Config\InstallStorage;
use Drupal\Core\Config\StorageComparer;
use Drupal\Core\Logger\LoggerChannelTrait;
use Drupal\Core\Messenger\MessengerTrait;
use Drupal\Core\StringTranslation\StringTranslationTrait;
use Drupal\config\StorageReplaceDataWrapper;

/**
 * Enforces configuration.
 */
class EnforcedConfig {

  use StringTranslationTrait, MessengerTrait, LoggerChannelTrait;

  const CONFIG_PREFIX = 'config_enforce.enforced_configs';

  const CONFIG_ENFORCE_OFF = 0;
  const CONFIG_ENFORCE_NOSUBMIT = 10;
  const CONFIG_ENFORCE_READONLY = 20;

  /**
   * Return whether a given config object is being enforced.
   */
  public static function isEnforced(string $config_name) {
    return !is_null(self::getEnforcedConfig($config_name));
  }

  /**
   * Return the currently configured target module for a given config object.
   */
  public static function getTargetModule(string $config_name) {
    return self::getEnforcedConfig($config_name)['target_module'];
  }

  /**
   * Return the currently configured config directory for a given config object.
   */
  public static function getConfigDirectory(string $config_name) {
    return self::getEnforcedConfig($config_name)['config_directory'];
  }

  /**
   * Return the currently configured path of the config object.
   */
  public static function getConfigFilePath(string $config_name) {
    return self::getEnforcedConfig($config_name)['config_file_path'];
  }

  /**
   * Return the URI of the host form, if one is available.
   */
  public static function getConfigFormUri(string $config_name) {
    return self::getEnforcedConfig($config_name)['config_form_uri'];
  }

  /**
   * Return a list of available enforcement levels.
   */
  public static function getEnforcementLevels() {
    return [
      self::CONFIG_ENFORCE_OFF => t('Allow form & API updates.'),
      self::CONFIG_ENFORCE_NOSUBMIT => t('Allow only API updates.'),
      self::CONFIG_ENFORCE_READONLY => t('Read-only, no updates.'),
    ];
  }

  /**
   * Return the currently configured enforcement level for the config object.
   */
  public static function getEnforcementLevel(string $config_name) {
    return self::getEnforcedConfig($config_name)['enforcement_level'];
  }

  /**
   * Return a human readable enforcement level for the config object.
   */
  public static function getEnforcementLevelLabel(string $config_name) {
    return self::getEnforcementLevels()[self::getEnforcementLevel($config_name)];
  }

  /**
   * Return this module's stored config.
   */
  public static function getConfig(string $module) {
    return \Drupal::config(self::CONFIG_PREFIX . '.' . $module);
  }

  /**
   * Return this module's configuration for a given config object.
   */
  public static function getEnforcedConfig(string $config_name) {
    $enforced_configs = self::getEnforcedConfigs();

    // If a config isn't present, that config is not enforced.
    if (!array_key_exists($config_name, $enforced_configs)) return NULL;

    return $enforced_configs[$config_name];
  }

  /**
   * Return this module's configuration for all config objects, except those
   * from disabled modules.
   */
  public static function getEnforcedConfigs() {
    $configs = &drupal_static(__FUNCTION__);
    if (!isset($configs)) {
      foreach (self::getAllModules() as $module => $label) {
        foreach (self::getConfig($module)->get() as $encoded_name => $info) {
          if ($encoded_name == '_core') continue;
          $configs[self::decode($encoded_name)] = $info;
        }
      }
    }
    return $configs;
  }

  /**
   * Import all enforced config registry configs from disk.
   */
  protected function readEnforcedConfigs() {
    foreach (\Drupal::moduleHandler()->getModuleDirectories() as $module => $module_path) {
      $config_file_path = $module_path . '/';
      $config_file_path .= InstallStorage::CONFIG_INSTALL_DIRECTORY . '/';
      $config_file_path .= self::CONFIG_PREFIX . '.' . $module . '.yml';
      if (!file_exists($config_file_path)) continue;
      self::importConfig($config_file_path);
    }
    drupal_static_reset('getEnforcedConfigs');
  }

  /**
   * Return a list of modules.
   */
  public static function getAllModules() {
    $modules = &drupal_static(__FUNCTION__);
    if (!isset($modules)) {
      $moduleHandler = \Drupal::moduleHandler();
      foreach ($moduleHandler->getModuleList() as $name => $module) {
        $modules[$name] = $moduleHandler->getName($name);
      }
      asort($modules);
    }
    return $modules;
  }

  /**
   * Consistently encode config names, so that they can be used as array keys when saved.
   *
   * @see decode().
   */
  public static function encode($config_name) {
    return str_replace('.', '__', $config_name);
  }

  /**
   * Consistently decode config names.
   *
   * @see encode().
   */
  public static function decode($encoded_config_name) {
    return str_replace('__', '.', $encoded_config_name);
  }

  /**
   * Re-import all read-only config.
   */
  public static function enforceConfigs() {
    (new static())->readEnforcedConfigs();
    $registries = EnforcedConfigRegistry::getAllConfigNames();

    foreach (self::getEnforcedConfigs() as $config_name => $setting) {
      // We just reloaded the registry configs from disk, so it's safe to skip them here.
      if (in_array($config_name, $registries)) continue;
      // Only re-import read-only configs.
      if ($setting['enforcement_level'] < self::CONFIG_ENFORCE_READONLY) continue;
      // Only re-import configs that have changed from what's on disk.
      if (!self::configHasChanged($config_name, $setting['hash'])) continue;
      self::importConfig($setting['config_file_path']);
      // @TODO log each import?
    }
  }

  /**
   * Determine whether config in active storage is different from what is on disk.
   */
  protected static function configHasChanged(string $config_name, string $hash) {
    $config = \Drupal::config($config_name)->get();
    // @TODO: find a better way to ignore uuid in general case, rather than special case system.site
    if ($config_name != 'system.site') unset($config['uuid']);
    $data = (new InstallStorage())->encode($config);
    return $hash != Crypt::hashBase64($data);
  }

  /**
   * Imports a config item from the given filename.
   *
   * @param string $filename
   *   The filename to import from with the Drupal-relative path.
   */
  protected static function importConfig($filename) {
    $contents = @file_get_contents($filename);
    $data = (new InstallStorage())->decode($contents);
    $config_name = basename($filename, '.yml');

    self::handleSpecialCaseConfig($config_name, $data);

    $config_storage = \Drupal::service('config.storage');
    $source_storage = new StorageReplaceDataWrapper($config_storage);
    $source_storage->replaceData($config_name, $data);
    $storage_comparer = new StorageComparer($source_storage, $config_storage);

    $storage_comparer->createChangelist();

    // @TODO log whether config has changed?

    return self::getConfigImporter($storage_comparer)->import();
  }

  /**
   * Return an instantiated ConfigImporter.
   */
  protected static function getConfigImporter($storage_comparer) {
    // @TODO: simplify this when
    // https://www.drupal.org/project/drupal/issues/3123491 is fixed.
    $config_importer = new ConfigImporter(
      $storage_comparer,
      \Drupal::service('event_dispatcher'),
      \Drupal::service('config.manager'),
      \Drupal::service('lock.persistent'),
      \Drupal::service('config.typed'),
      \Drupal::service('module_handler'),
      \Drupal::service('module_installer'),
      \Drupal::service('theme_handler'),
      \Drupal::service('string_translation')
    );

    return $config_importer;
  }

  /**
   * Alter data or otherwise handle config special cases.
   */
  protected static function handleSpecialCaseConfig($config_name, &$data) {
    switch ($config_name) {
      case 'system.site':
        // Site UUID gets validated, so we need to make sure it'll match.
        // @TODO log this?
        $data['uuid'] = \Drupal::config('system.site')->get('uuid');
        break;
      default:
        return;
    }
  }

}
